from json import JSONEncoder
from typing import Optional, Dict, Any, List, Union

import requests


class SnapshotSearchNVAPIV3Result:
    def __init__(self, query: Dict[str, str], response: Union[requests.Response, List['SnapshotSearchNVAPIV3Result']]):
        self._json: Optional[Dict[str, Any]] = None
        self.query = query

        if isinstance(response, requests.Response):
            self._json: Optional[Dict[str, Any]] = response.json()
            return

        results: List['SnapshotSearchNVAPIV3Result'] = response

        if len(results) < 1:
            raise Exception("空のリストが渡されました")
        data = {'totalCount': results[0].data()['totalCount'], 'items': []}
        meta = results[0].meta()
        for result in results:
            data['items'].extend(result.data()['items'])
        self._json: Optional[Dict[str, Any]] = {"meta": meta, "data": data}

    def json(self) -> Dict[str, Any]:
        return self._json

    def meta(self) -> Dict[str, Any]:
        if 'meta' not in self.json():
            return {}
        return self.json()['meta']

    def data(self) -> Dict[str, Any]:
        if 'data' not in self.json():
            return {}
        return self.json()['data']

    def items(self) -> List[Dict[str, Any]]:
        if 'items' not in self.data():
            return []
        return self.data()['items']

    def status(self) -> int:
        return self.meta()['status']

    def total_count(self) -> int:
        return self.data()['totalCount']

    def text(self) -> str:
        return JSONEncoder().encode(self.json())



from nicovideo_api_client.api.v2.targets import SnapshotSearchAPIV2TagsExact, SnapshotSearchAPIV2Keywords


class SnapshotSearchAPIV2:
    @staticmethod
    def tags_exact() -> 'SnapshotSearchAPIV2TagsExact':
        return SnapshotSearchAPIV2TagsExact()

    @staticmethod
    def keywords() -> 'SnapshotSearchAPIV2Keywords':
        return SnapshotSearchAPIV2Keywords()



from typing import Dict

from nicovideo_api_client.nvapi.v3.limit import SnapshotSearchNVAPIV3Limit


class SnapshotSearchNVAPIV3SortKey:
    def __init__(self, endpoint: str, query: Dict[str, str]):
        self.endpoint = endpoint
        self.query: Dict[str, str] = query

    def sort_order(self, asc: bool) -> SnapshotSearchNVAPIV3Limit:
        self.query['sortOrder'] = 'asc' if asc else 'desc'
        return SnapshotSearchNVAPIV3Limit(self.endpoint, self.query)



import urllib
from typing import Dict, List
from urllib.parse import urlencode, unquote_plus

import math
import requests
import time

from nicovideo_api_client.nvapi.v2.result import SnapshotSearchNVAPIV2Result


class SnapshotSearchNVAPIV2Request:

    def __init__(self, endpoint: str, query: Dict[str, str], limit: int):
        self.endpoint = endpoint
        self.query: Dict[str, str] = query
        self.limit = limit

    @staticmethod
    def get(url: str):
        return requests.get(url, headers={'X-Frontend-Id': '6'})

    def request(self) -> SnapshotSearchNVAPIV2Result:
        if self.limit <= 100:
            self.query["pageSize"] = str(self.limit)
        else:
            self.query["pageSize"] = "100"

        results: List[SnapshotSearchNVAPIV2Result] = [SnapshotSearchNVAPIV2Result(
            self.query,
            self.get(self.build_url())
        )]

        total_count = int(results[0].total_item_count())

        if total_count <= self.limit:
            self.limit = total_count

        for pos in range(1, math.ceil(self.limit / 100)):
            self.query["page"] = str(pos * 100)
            if self.limit < (pos + 1) * 100:
                self.query["pageSize"] = str(self.limit % 100)
            tmp = SnapshotSearchNVAPIV2Result(self.query, self.get(self.build_url()))
            while "meta" not in tmp.json() or tmp.status() != 200:
                print("Connection Failed!")
                time.sleep(1.5)
                tmp = SnapshotSearchNVAPIV2Result(self.query, self.get(self.build_url()))
            results.append(tmp)

        return SnapshotSearchNVAPIV2Result(self.query, results)

    def build_url(self, decode: bool = False) -> str:
        query = urlencode(self.query)
        if decode:
            query = urllib.parse.unquote_plus(query)
        return self.endpoint + '?' + unquote_plus(query)



from datetime import datetime
from typing import Dict, List, Optional, Union


class JsonFilterTerm:
    def __init__(self):
        self.json: Dict[str, Union[str, bool, int]] = {}

    @staticmethod
    def set_range_view(
            start: Optional[int] = None,
            end: Optional[int] = None,
            include_lower: bool = True,
            include_upper: bool = True,
    ) -> 'JsonFilterTerm':
        term: JsonFilterTerm = JsonFilterTerm()
        if start is None and end is None:
            raise Exception("上限も下限も指定されていません")
        json_: Dict[str, Union[str, bool, int]] = {
            "type": "range",
            "field": "viewCounter"
        }
        if start is not None:
            json_["from"] = start
        if end is not None:
            json_["to"] = end
        if include_lower:
            json_["include_lower"] = include_lower
        if include_upper:
            json_["include_upper"] = include_upper
        term.json = json_
        return term

    @staticmethod
    def set_range_time(
            start: Optional[datetime] = None,
            end: Optional[datetime] = None,
            include_lower: bool = True,
            include_upper: bool = True,
    ) -> 'JsonFilterTerm':
        term: JsonFilterTerm = JsonFilterTerm()
        if start is None and end is None:
            raise Exception("開始時刻も終了時刻も指定されていません")
        json_: Dict[str, Union[str, bool]] = {
            "type": "range",
            "field": "startTime"
        }
        if start is not None:
            json_["from"] = start.strftime('%Y-%m-%dT%H:%M:%S+09:00')
        if end is not None:
            json_["to"] = end.strftime('%Y-%m-%dT%H:%M:%S+09:00')
        if include_lower:
            json_["include_lower"] = include_lower
        if include_upper:
            json_["include_upper"] = include_upper
        term.json = json_
        return term


TypeOp = Dict[str, Union[str, bool, int, List['TypeOp'], 'TypeOp']]


class JsonFilterOperator:
    def __init__(self, json_: TypeOp):
        self.json: TypeOp = json_

    @staticmethod
    def unit(term: 'JsonFilterTerm') -> 'JsonFilterOperator':
        return JsonFilterOperator(term.json)

    @staticmethod
    def not_(
            term: Union['JsonFilterTerm', 'JsonFilterOperator']
    ):
        json_: TypeOp = {
            "type": "not",
            "filter": term.json
        }
        return JsonFilterOperator(json_)

    @staticmethod
    def _binary(
            op_name: str,
            left: Union['JsonFilterTerm', 'JsonFilterOperator'],
            right: Union['JsonFilterTerm', 'JsonFilterOperator']
    ):
        json_: TypeOp = {
            "type": op_name,
            "filters": []
        }
        json_["filters"].append(left.json)
        json_["filters"].append(right.json)
        return JsonFilterOperator(json_)

    @staticmethod
    def or_(
            left: Union['JsonFilterTerm', 'JsonFilterOperator'],
            right: Union['JsonFilterTerm', 'JsonFilterOperator']
    ) -> 'JsonFilterOperator':
        return JsonFilterOperator._binary("or", left, right)

    @staticmethod
    def and_(
            left: Union['JsonFilterTerm', 'JsonFilterOperator'],
            right: Union['JsonFilterTerm', 'JsonFilterOperator']
    ) -> 'JsonFilterOperator':
        return JsonFilterOperator._binary("and", left, right)



import statistics
from json import JSONEncoder
from typing import Optional, Dict, Any, List, Union

import math
import requests

from nicovideo_api_client.constants import FieldType


class SnapshotSearchAPIV2Result:
    def __init__(self, query: Dict[str, str], response: Union[requests.Response, List['SnapshotSearchAPIV2Result']]):
        self._fields: List[str] = query["fields"].split(",") if "fields" in query else []
        self._json: Optional[Dict[str, Any]] = None

        if isinstance(response, requests.Response):
            self._json: Optional[Dict[str, Any]] = response.json()
            return

        results: List['SnapshotSearchAPIV2Result'] = response

        if len(results) < 1:
            raise Exception("空のリストが渡されました")
        data = []
        meta = results[0].json()["meta"]
        for result in results:
            data.extend(result.json()["data"])
        self._json: Optional[Dict[str, Any]] = {"meta": meta, "data": data}

    def json(self) -> Dict[str, Any]:
        return self._json

    def meta(self) -> Dict[str, Any]:
        if 'meta' not in self.json():
            return {}
        return self.json()['meta']

    def data(self) -> List[Dict[str, Any]]:
        if 'data' not in self.json():
            return []
        return self.json()['data']

    def sum_view_counter(self) -> int:
        if FieldType.VIEW_COUNTER.value not in self._fields:
            raise Exception("フィールドに viewCounter が指定されていません")
        return sum(list(map(lambda x: x[FieldType.VIEW_COUNTER.value], self.data())))

    def avg_view_counter(self) -> int:
        if FieldType.VIEW_COUNTER.value not in self._fields:
            raise Exception("フィールドに viewCounter が指定されていません")
        return statistics.mean(list(map(lambda x: x[FieldType.VIEW_COUNTER.value], self.data())))

    def center_view_counter(self) -> int:
        if FieldType.VIEW_COUNTER.value not in self._fields:
            raise Exception("フィールドに viewCounter が指定されていません")
        return self.data()[math.floor(self.total_count() / 2)][FieldType.VIEW_COUNTER.value]

    def status(self) -> int:
        return self.meta()['status']

    def total_count(self) -> int:
        return self.meta()['totalCount']

    def text(self) -> str:
        return JSONEncoder().encode(self.json())



from typing import Dict

from nicovideo_api_client.api.v2.field import SnapshotSearchAPIV2Fields


class SnapshotSearchAPIV2TagsExact:
    def __init__(self):
        self._query: Dict[str, str] = {'targets': 'tagsExact'}

    def query(self, keyword: str) -> SnapshotSearchAPIV2Fields:
        self._query['q'] = keyword
        return SnapshotSearchAPIV2Fields(self._query)


class SnapshotSearchAPIV2Keywords:
    def __init__(self):
        self._query: Dict[str, str] = {'targets': 'title,description,tags'}

    def query(self, keyword: str) -> SnapshotSearchAPIV2Fields:
        self._query['q'] = keyword
        return SnapshotSearchAPIV2Fields(self._query)



from typing import Dict

from nicovideo_api_client.nvapi.v2.request import SnapshotSearchNVAPIV2Request


class SnapshotSearchNVAPIV2Mylists:
    def __init__(self, endpoint: str, mylist_id: str):
        self.endpoint = endpoint + mylist_id
        self.query: Dict[str, str] = {}

    def limit(self, limit: int) -> SnapshotSearchNVAPIV2Request:
        return SnapshotSearchNVAPIV2Request(self.endpoint, self.query, limit)



from typing import Union

from nicovideo_api_client.constants import END_POINT_URL_NVV2
from nicovideo_api_client.nvapi.v2.mylists import SnapshotSearchNVAPIV2Mylists


class SnapshotSearchNVAPIV2:

    @staticmethod
    def mylists(mylist_id: Union[int, str]) -> SnapshotSearchNVAPIV2Mylists:
        if isinstance(mylist_id, int):
            mylist_id_str = f'mylists/{mylist_id}'
        else:
            mylist_id_str = mylist_id
        return SnapshotSearchNVAPIV2Mylists(END_POINT_URL_NVV2, mylist_id_str)



from typing import Dict

from nicovideo_api_client.api.v2.request import SnapshotSearchAPIV2Request


class SnapshotSearchAPIV2Limit:

    def __init__(self, query: Dict[str, str]):
        self._query: Dict[str, str] = query

    def limit(self, limit: int = 4) -> SnapshotSearchAPIV2Request:
        return SnapshotSearchAPIV2Request(self._query, limit)



from typing import Dict

from nicovideo_api_client.nvapi.v3.request import SnapshotSearchNVAPIV3Request


class SnapshotSearchNVAPIV3Limit:
    def __init__(self, endpoint: str, query: Dict[str, str]):
        self.endpoint = endpoint
        self.query: Dict[str, str] = query

    def limit(self, limit: int) -> SnapshotSearchNVAPIV3Request:
        return SnapshotSearchNVAPIV3Request(self.endpoint, self.query, limit)



from typing import Dict, List, Tuple
from urllib.parse import urlencode, unquote_plus

import math
import requests
import time

from nicovideo_api_client.api.v2.result import SnapshotSearchAPIV2Result
from nicovideo_api_client.constants import END_POINT_URL_V2


class SnapshotSearchAPIV2Request:

    def __init__(self, query: Dict[str, str], limit: int):
        self._query: Dict[str, str] = query
        self._limit = limit

    def request(self, timeout: float = 400.0) -> SnapshotSearchAPIV2Result:
        if self._limit <= 100:
            self._query["_limit"] = str(self._limit)
        else:
            self._query["_limit"] = "100"

        results: List[SnapshotSearchAPIV2Result] = [SnapshotSearchAPIV2Result(
            self._query,
            requests.get(self.build_url())
        )]

        total_count = int(results[0].total_count())

        if total_count <= self._limit:
            self._limit = total_count

        total_time = 0.0

        for pos in range(1, math.ceil(self._limit / 100)):
            self._query["_offset"] = str(pos * 100)
            if self._limit < (pos + 1) * 100:
                self._query["_limit"] = str(self._limit % 100)

            while True:
                (response, response_time) = self._request(timeout)
                total_time += response_time
                if total_time > timeout:
                    raise TimeoutError("通信がタイムアウトしました")
                if "meta" in response.json() or response.status() == 200:
                    break
                print("Connection Failed!")
                time.sleep(1.5)

            results.append(response)
        return SnapshotSearchAPIV2Result(self._query, results)

    def _request(self, timeout: float) -> Tuple[SnapshotSearchAPIV2Result, float]:
        response = requests.get(self.build_url(), timeout=(timeout / 4, timeout * 3 / 4))
        total_time = response.elapsed.total_seconds()
        tmp = SnapshotSearchAPIV2Result(self._query, response)
        return tmp, total_time

    def build_url(self, decode: bool = True) -> str:
        query = urlencode(self._query)
        if decode:
            query = unquote_plus(query)
        return END_POINT_URL_V2 + '?' + query



from typing import Dict

from nicovideo_api_client.constants import SortKeyType
from nicovideo_api_client.nvapi.v3.sortkey import SnapshotSearchNVAPIV3SortKey


class SnapshotSearchNVAPIV3Users:
    def __init__(self, endpoint: str, user_id: int):
        self.endpoint = endpoint + f'users/{user_id}/videos'
        self.query: Dict[str, str] = {}

    def sort_key(self, sort_key_type: SortKeyType) -> SnapshotSearchNVAPIV3SortKey:
        self.query['sortKey'] = sort_key_type.value
        return SnapshotSearchNVAPIV3SortKey(self.endpoint, self.query)



from nicovideo_api_client.constants import END_POINT_URL_NVV3
from nicovideo_api_client.nvapi.v3.users import SnapshotSearchNVAPIV3Users


class SnapshotSearchNVAPIV3:

    @staticmethod
    def users(user_id: int) -> SnapshotSearchNVAPIV3Users:
        return SnapshotSearchNVAPIV3Users(END_POINT_URL_NVV3, user_id)
